//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//
//
// Use of this sample source code is subject to the terms of the Microsoft
// license agreement under which you licensed this sample source code. If
// you did not accept the terms of the license agreement, you are not
// authorized to use this sample source code. For the terms of the license,
// please see the license agreement between you and Microsoft or, if applicable,
// see the LICENSE.RTF on your install media or the root of your tools installation.
// THE SAMPLE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES OR INDEMNITIES.
//
//
//

//
// Copyright (C) 2011, UNITEC Co., LTD. All Rights Reserved.
// THIS SOURCE CODE, AND ITS USE AND DISTRIBUTION, IS SUBJECT TO THE TERMS
// AND CONDITIONS OF THE APPLICABLE LICENSE AGREEMENT
//

#include <windows.h>
#include <windev.h>
#include <svsutil.hxx>
#include <tchstreamddsi.h>

#include <mtouch.h>
#include "mtouch_hid.h"



#define REG_GUID_FORMAT_W   L"{%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x}"

// Debug zones
#ifndef SHIP_BUILD

#undef ZONE_ERROR
#undef ZONE_WARNING
#undef ZONE_INIT
#undef ZONE_FUNCTION
#undef ZONE_INFO
#undef ZONE_TIPSTATE
#undef ZONE_USAGES

#define ZONE_ERROR          DEBUGZONE(0)
#define ZONE_WARN           DEBUGZONE(1)
#define ZONE_INIT           DEBUGZONE(2)
#define ZONE_FUNCTION       DEBUGZONE(3)
#define ZONE_INFO           DEBUGZONE(4)
#define ZONE_TIPSTATE       DEBUGZONE(5)
#define ZONE_USAGES         DEBUGZONE(6)

extern DBGPARAM dpCurSettings;

#else

#endif

static TCH_BSP_PROPS g_HidTouchProps;
static DWORD g_dwIstPrio = 0;

#ifdef DEBUG

static
void
ValidateHidTouch(
    PHID_TOUCH pHidTouch
    );

#else

#define ValidateHidTouch(ptr)

#endif // DEBUG


// Dll entry function.
extern "C" 
BOOL
DllEntry(
    HANDLE hDllHandle,
    DWORD dwReason, 
    LPVOID lpReserved
    )
{
    UNREFERENCED_PARAMETER(lpReserved);
    
    switch (dwReason) {
        case DLL_PROCESS_ATTACH:
            DEBUGREGISTER((HINSTANCE)hDllHandle);
            DEBUGMSG(ZONE_INIT, (_T("%S: Attach\r\n"), __FUNCTION__));
            DisableThreadLibraryCalls((HMODULE) hDllHandle);
            break;
            
        case DLL_PROCESS_DETACH:
            DEBUGMSG(ZONE_INIT, (_T("%S: Detach\r\n"), __FUNCTION__));
            break;
            
        default:
            break;
    }
    
    return TRUE ;
}

static 
HANDLE 
SetHidAttachEvent(void)
{
	HANDLE hHidAttach;

	hHidAttach = CreateEvent(NULL, FALSE, FALSE, TCHHID_HID_ATTACH_NAME);
	if (hHidAttach == NULL)
	{
    	DEBUGMSG(ZONE_ERROR, (_T("%S: CreateEvent error = 0x%xr\n"), __FUNCTION__, GetLastError()));
        goto EXIT;
	}
	SetEvent(hHidAttach);

EXIT:
	return hHidAttach;	
}

static 
void 
WaitForHidAttach(void)
{
	HANDLE hHidAttach = CreateEvent(NULL, FALSE, FALSE, TCHHID_HID_ATTACH_NAME);
	if (hHidAttach == NULL)
	{
    	DEBUGMSG(ZONE_ERROR, (_T("%S: CreateEvent error = 0x%xr\n"), __FUNCTION__, GetLastError()));
		goto EXIT;
	}

	if (GetLastError() != ERROR_ALREADY_EXISTS)
	{
		while(WAIT_OBJECT_0 != WaitForSingleObject(hHidAttach, 1))
		{
			Sleep(100);
		}

	}

	CloseHandle(hHidAttach);
EXIT:
	return;	
}

// BSP func: Enable the interrupts
static
BOOL 
HidTouchInterruptEnable(LPVOID lpBspContext)
{
    UNREFERENCED_PARAMETER(lpBspContext);
	// Nothing to do for the controller.
	
	return TRUE;
}

// BSP func: Disable the interrupts
static
void 
HidTouchInterruptDisable(LPVOID lpBspContext)
{
    UNREFERENCED_PARAMETER(lpBspContext);
	// Nothing to do for the controller.
}

// BSP func: Clear the interrupts
static 
void 
HidTouchInterruptDone(LPVOID lpBspContext)
{
    UNREFERENCED_PARAMETER(lpBspContext);
	// Nothing to do for the controller.
}

// BSP func: Handle the power status
static 
void 
HidTouchPowerHandler(LPVOID lpBspContext, BOOL bOff)
{
    UNREFERENCED_PARAMETER(lpBspContext);
	// Nothing to do for the controller.
}

// BSP func: Get the samples which has been already queued 
static
BOOL 
HidTouchGetSample(
    LPVOID lpBspContext,
	TCH_BSP_INPUT_INFO *pInput,
	INT *count
	)
{

	PTCH_STREAM pTchStream = (PTCH_STREAM)lpBspContext;
	TCHHID_INFO info = {0};
    DWORD dwRead;
    DWORD dwFlags;
	BOOL fRet = FALSE;

	if (!pTchStream)
		return FALSE;

	if (!ReadMsgQueue(pTchStream->hMsgQueueReader, &info, sizeof(info), &dwRead, 0, &dwFlags))
	{
		goto exit;
	}

	if (dwRead != sizeof(info)) {
		info.dwCount = 0;
	}
	
	for (DWORD i = 0; i < info.dwCount; i++) {
		pInput[i] = info.pBspInput[i];
	}
	
	fRet = TRUE; 
exit:
	*count = info.dwCount;
	return fRet;
}

// BSP func: Get the controller properties
static 
BOOL 
HidTouchGetProps (
    LPVOID lpBspContext,
	TCH_BSP_PROPS *pProps
    )
{
	BOOL fRet = FALSE;

	// Check that Touch-Panel is attached and wait for getting ready
	WaitForHidAttach();

	// Then, copy the controller properties into the upper's buffer
	if (pProps)
	{
		pProps->dwMaxContact = g_HidTouchProps.dwMaxContact;
		fRet = TRUE;
	}

	return fRet;
}

// Create P2P Message Queue between IST and PDD-called function
static 
HANDLE 
CreateTouchMsgQueue(
	BOOL read
	)
{
	HANDLE hRet;
	MSGQUEUEOPTIONS * popts;
	MSGQUEUEOPTIONS reader_opts = {
		sizeof(reader_opts),
		0,
		TCHHID_MSG_QUEUE_MAX,
		sizeof(TCHHID_INFO),
		TRUE
	};
	MSGQUEUEOPTIONS writer_opts = {
		sizeof(writer_opts),
		MSGQUEUE_ALLOW_BROKEN,
		TCHHID_MSG_QUEUE_MAX,
		sizeof(TCHHID_INFO),
		FALSE
	};

	if (read) 
	{
		popts = &reader_opts;
	}
	else
	{
		popts = &writer_opts;
	}

	hRet = CreateMsgQueue(TCHHID_MSG_QUEUE_NAME, popts);
	if (hRet == NULL)
	{
		DEBUGMSG(ZONE_ERROR, (L"%S: Could not create the message queue. Error=%u\r\n", 
					__FUNCTION__, GetLastError()));
	}
	return hRet;
}

// the Initializes Function which be called by PDD
extern "C" 
HANDLE 
TchBsp_Init(
	LPVOID pBspContext, 
	TCH_BSP_INTERFACE_INFO *pBspIfc,
	DWORD *pdwMaxContact,
	DWORD dwISTPriority
	)
{
	PTCH_STREAM pTchStream;

    RETAILMSG(1, (_T("%S: Driver Version = %s\r\n"), __FUNCTION__, TCH_DRIVER_VERSION));

    pTchStream = (PTCH_STREAM) LocalAlloc(LPTR, sizeof(TCH_STREAM));
    if (pTchStream == NULL) {
        DEBUGMSG(ZONE_ERROR, (_T("%S: LocalAlloc error:%d\r\n"), __FUNCTION__, GetLastError()));
        return NULL;
    }

	pTchStream->hMsgQueueReader = CreateTouchMsgQueue(TRUE);
    if (pTchStream->hMsgQueueReader == NULL)
	{
		LocalFree(pTchStream);
        return NULL;
    }

	if (pBspIfc) 
	{
		pBspIfc->pfnInterruptEnable = HidTouchInterruptEnable;
		pBspIfc->pfnInterruptDisable = HidTouchInterruptDisable;
		pBspIfc->pfnInterruptDone = HidTouchInterruptDone;
		pBspIfc->pfnPowerHandler = HidTouchPowerHandler;
		pBspIfc->pfnGetSample = HidTouchGetSample;
		pBspIfc->pfnGetProps = HidTouchGetProps;
	}

	if (pBspContext)
	{
		*((PTCH_STREAM *)pBspContext) = pTchStream;
	}

	if (pdwMaxContact)
	{
		*pdwMaxContact = TCHHID_MAX_CONTACT;
	}

	g_dwIstPrio = dwISTPriority;

	return pTchStream->hMsgQueueReader;
}


static
NTSTATUS
GetGenericUsageValue(
    PHID_TOUCH pHidTouch,
	USHORT link, 
    USAGE usage,
    PCHAR pbHidPacket,
    DWORD cbHidPacket,
    PLONG pValue
    )
{
    NTSTATUS status;
    
    PREFAST_DEBUGCHK(pHidTouch != NULL);
    DEBUGCHK(pbHidPacket != NULL);
    
    status = HidP_GetScaledUsageValue(
        HidP_Input,
        HID_USAGE_PAGE_GENERIC,
        link,
        usage,
        pValue,
        pHidTouch->phidpPreparsedData,
        pbHidPacket,
        cbHidPacket);
    if (status == HIDP_STATUS_BAD_LOG_PHY_VALUES) {
        // Bad physical max/min detected.
        DEBUGMSG(ZONE_ERROR, (_T("%S: Bad physical max/min for usage 0x%x\r\n"),
            __FUNCTION__, usage));
        *pValue = 0;
    }

    return status;
}

static
NTSTATUS
GetDigitizerUsageValue(
	HIDP_REPORT_TYPE ReportType,
	USHORT link, 
    USAGE usage,
	PHIDP_PREPARSED_DATA PreparsedData,
    PCHAR pbHidPacket,
    DWORD cbHidPacket,
    PULONG pValue
    )
{
    NTSTATUS status;
    
    DEBUGCHK(pbHidPacket != NULL);
    PREFAST_DEBUGCHK(pValue != NULL);
    
    status = HidP_GetUsageValue(
        ReportType,
        HID_USAGE_PAGE_DIGITIZER,
        link,
        usage,
        pValue,
        PreparsedData,
        pbHidPacket,
        cbHidPacket);
    if (status == HIDP_STATUS_BAD_LOG_PHY_VALUES) {
        // Bad physical max/min detected.
        DEBUGMSG(ZONE_ERROR, (_T("%S: Bad physical max/min for usage 0x%x\r\n"),
            __FUNCTION__, usage));
        *pValue = 0;
    }

    return status;
}

static
TOUCH_PANEL_SAMPLE_FLAGS
GetTouchSampleFlag(
    PHID_TOUCH pHidTouch,
	USHORT link, 
    PCHAR pbHidPacket,
    DWORD cbHidPacket
    )
{
    NTSTATUS status;
    ULONG uCurrUsages;
	TOUCH_PANEL_SAMPLE_FLAGS flags = 0;
    
    PREFAST_DEBUGCHK(pHidTouch != NULL);
    DEBUGCHK(pbHidPacket != NULL);
    
	uCurrUsages = pHidTouch->dwMaxUsages;

	// Get the contact flags from each collections
	status = HidP_GetUsages(
			HidP_Input,
			HID_USAGE_PAGE_DIGITIZER,
			link,
			pHidTouch->puUsages,
			&uCurrUsages, // IN OUT parameter
			pHidTouch->phidpPreparsedData,
			pbHidPacket,
			cbHidPacket);
	if (status != HIDP_STATUS_SUCCESS) {
		DEBUGMSG(ZONE_ERROR, (_T("%S: collection = %d starts = 0x%08x \r\n"), __FUNCTION__, 1, status));
		goto exit;
	}

	flags = TouchSampleValidFlag;
	for (UINT i = 0; i < uCurrUsages; i++) {
		USAGE Usage = pHidTouch->puUsages[i];
		switch (Usage) {
			case HID_USAGE_DIGITIZER_TIP_SWITCH:
				flags |= TouchSampleDownFlag;
				break;
			case HID_USAGE_DIGITIZER_IN_RANGE:
				break;
			case HID_USAGE_DIGITIZER_CONFIDENCE:
				break;
			default:
				DEBUGMSG(ZONE_ERROR, (_T("%S: Unknown Page =%d \r\n"), __FUNCTION__, Usage));
				break;

		}
	}

exit:
	return flags;
}

static
void
HidGetXY(
    PHID_TOUCH pHidTouch,
    PCHAR pbHidPacket,
    DWORD cbHidPacket,
    LONG *x,
    LONG *y
    )
{
    PREFAST_DEBUGCHK(pHidTouch != NULL);
    DEBUGCHK(pbHidPacket != NULL);

    ValidateHidTouch(pHidTouch);

    // Get the x-y coordinates
    GetGenericUsageValue(pHidTouch, 1, HID_USAGE_GENERIC_X, 
            pbHidPacket, cbHidPacket, x);
    GetGenericUsageValue(pHidTouch, 1, HID_USAGE_GENERIC_Y, 
            pbHidPacket, cbHidPacket, y);
}

static
void
HidGetTouchFlags(
    PHID_TOUCH pHidTouch,
    PCHAR pbHidPacket,
    DWORD cbHidPacket,
    TOUCH_PANEL_SAMPLE_FLAGS *touchFlags
    )
{
    TOUCH_PANEL_SAMPLE_FLAGS flags;

    PREFAST_DEBUGCHK(pHidTouch != NULL);
    DEBUGCHK(pbHidPacket != NULL);

    ValidateHidTouch(pHidTouch);

    // Get the touch flags
    flags = GetTouchSampleFlag(pHidTouch, 1, pbHidPacket, cbHidPacket);
    if (!(flags & TouchSampleValidFlag))
    {
        return;
    }
    *touchFlags = flags;
}

static void HidGetContactID(
    PHID_TOUCH pHidTouch,
    PCHAR pbHidPacket,
    DWORD cbHidPacket,
    DWORD *dwContactID
    )
{
	NTSTATUS status;
    DWORD dwID;
    
    PREFAST_DEBUGCHK(pHidTouch != NULL);
    DEBUGCHK(pbHidPacket != NULL);

    ValidateHidTouch(pHidTouch);

    // Get the contact ID
    status = GetDigitizerUsageValue(HidP_Input, 1, HID_USAGE_DIGITIZER_CONTACT_ID, 
            pHidTouch->phidpPreparsedData, pbHidPacket, cbHidPacket, &dwID);
    if (status != HIDP_STATUS_SUCCESS) 
    {
        return;
    }
    *dwContactID = dwID;
}

static
void
HidTouchPutSample(
    PHID_TOUCH pHidTouch,
	PTCHHID_INFO pInfo
	)
{
	if (!WriteMsgQueue(pHidTouch->hMsgQueueWriter, (LPVOID)pInfo, sizeof(TCHHID_INFO), 0, 0))
	{
		DWORD dwErr = GetLastError();
		if (ERROR_TIMEOUT != dwErr)
		{
			DEBUGMSG(ZONE_ERROR, (L"WriteMsgQueue: failed. Error=%u\r\n", dwErr));
		}
		else 
		{
			DEBUGMSG(ZONE_ERROR, (L"WriteMsgQueue: message queue is now full\r\n"));
		}
	}
}

// Get interrupt reports from the device. Thread exits when the device has
// been removed.
static
DWORD
WINAPI
TouchThreadProc(
    LPVOID lpParameter
    )
{
	PCHAR pbHidPacket = NULL;
	TCHHID_INFO info[2];	// Double buffer
	DWORD dwBufSwitch = 0;
	DWORD dwCount = 0;
	DWORD dwPreID = 0;
	DWORD dwCurrentID = 0;
	DWORD dwTimeout = INFINITE;

    PHID_TOUCH pHidTouch = (PHID_TOUCH) lpParameter;
    
    PREFAST_DEBUGCHK(pHidTouch != NULL);

    for (;;) 
    {
        if (g_dwIstPrio) 
        {
			CeSetThreadPriority(GetCurrentThread(), g_dwIstPrio);
			break;
        }
        Sleep(100);
    }
    
	// Allocate the input buffers
	info[0].pBspInput = (PTCH_BSP_INPUT_INFO)LocalAlloc(LPTR, 
							sizeof(TCH_BSP_INPUT_INFO) * g_HidTouchProps.dwMaxContact);
	info[1].pBspInput = (PTCH_BSP_INPUT_INFO)LocalAlloc(LPTR, 
							sizeof(TCH_BSP_INPUT_INFO) * g_HidTouchProps.dwMaxContact);
	if ((info[0].pBspInput == NULL) || (info[1].pBspInput == NULL))
	{
    	DEBUGMSG(ZONE_ERROR, (_T("%S: GLE=0x%08x\r\n"), __FUNCTION__, GetLastError()));
        goto EXIT;
	}
		
	// Allocate the HID buffers
    const DWORD cbMaxHidPacket = pHidTouch->hidpCaps.InputReportByteLength;
    pbHidPacket = (PCHAR) LocalAlloc(LMEM_FIXED, cbMaxHidPacket);
    if (pbHidPacket == NULL) {
        DEBUGMSG(ZONE_ERROR, (_T("%S: LocalAlloc error:%d\r\n"), __FUNCTION__, GetLastError()));
        goto EXIT;
    }

    for (;;) {
        ValidateHidTouch(pHidTouch);

        DWORD cbHidPacket;
            
        // Get an interrupt report from the device.
        DWORD dwErr = pHidTouch->pHidFuncs->lpGetInterruptReport(
            pHidTouch->hDevice,
            pbHidPacket,
            cbMaxHidPacket,
            &cbHidPacket,
            NULL,
            dwTimeout);
        if (dwErr != ERROR_SUCCESS) {
            if (dwErr == ERROR_TIMEOUT) {
                // disable timeout
                dwTimeout = INFINITE;

                // Write touch input to message queue binded to PDD 
                if( dwCount > 0 ) {
                    info[dwBufSwitch].dwCount = dwCount;
                    HidTouchPutSample(pHidTouch, &info[dwBufSwitch]);

                    // Reset Contact Count
                    dwCount = 0;
                    
                    // Switch Buffer
                    dwBufSwitch ^= 0x01;
                }
            }
            else {
                if ( (dwErr != ERROR_DEVICE_REMOVED) && (dwErr != ERROR_CANCELLED) ) {
                    DEBUGMSG(ZONE_ERROR, 
                        (_T("%S: GetInterruptReport returned unexpected error %u\r\n"),
                        __FUNCTION__, dwErr));
                }
                break;
            }
        }
        else {
            DEBUGCHK(cbHidPacket <= cbMaxHidPacket);
            // Set enable timeout
            dwTimeout = TCHHID_POLLING_TIMEOUT;

#if 0
    RETAILMSG(1, (_T("%x %x / %x %x / %x\r\n"), 
                (BYTE)pbHidPacket[1],
                (BYTE)pbHidPacket[2],
                (BYTE)pbHidPacket[7],
                (BYTE)pbHidPacket[8],
                (BYTE)pbHidPacket[13]));
#endif

            // Get Contact ID
            HidGetContactID(pHidTouch, pbHidPacket, cbHidPacket, &dwCurrentID);
            if( dwPreID >= dwCurrentID ) {
                // Write touch input to message queue binded to PDD 
                info[dwBufSwitch].dwCount = dwCount;
                HidTouchPutSample(pHidTouch, &info[dwBufSwitch]);
                
                // Reset Contact Count
                dwCount = 0;
                
                // Switch Buffer
                dwBufSwitch ^= 0x01;
            }
            
            // Set Contact ID
            info[dwBufSwitch].pBspInput[dwCount].dwID = dwCurrentID;

            // Set Touch Flags
            HidGetTouchFlags(
                pHidTouch, 
                pbHidPacket, 
                cbHidPacket, 
                &info[dwBufSwitch].pBspInput[dwCount].Flags);
            
            // Set XY Coordinate
            HidGetXY(
                pHidTouch, 
                pbHidPacket, 
                cbHidPacket, 
                &info[dwBufSwitch].pBspInput[dwCount].x, 
                &info[dwBufSwitch].pBspInput[dwCount].y);
            
            // Increment Contact Count
            dwCount++;
            
            // Save current FingerID
            dwPreID = dwCurrentID;
        }
    }

EXIT:
    if (pbHidPacket != NULL) LocalFree(pbHidPacket);
    if (info[0].pBspInput != NULL) LocalFree(info[0].pBspInput);
    if (info[1].pBspInput != NULL) LocalFree(info[1].pBspInput);
    DEBUGMSG(ZONE_FUNCTION, (_T("%S: Exiting thread\r\n"), __FUNCTION__));
    
    return 0;
}



// Free the memory used by pHidKbd
void
FreeHidTouch(
    PHID_TOUCH pHidTouch
    )
{
	if (pHidTouch->puUsages != NULL) LocalFree(pHidTouch->puUsages);
    if (pHidTouch != NULL) LocalFree(pHidTouch);    
}

#if 0 // delete by unitec k.kato
static 
void
AdvertiseHidTouchInterface(
    PHID_TOUCH pHidTouch,
    BOOL fAdd
    )
{
    union {
        BYTE rgbGuidBuffer[sizeof(GUID) + 4]; // +4 since scanf writes longs
        GUID guidBus;
    } u = { 0 };
    LPGUID pguidBus = &u.guidBus;
    LPCTSTR pszBusGuid = DRIVER_HID_TOUCH_GUID;
    WCHAR szName[50] = { 0 };
    DWORD dwIndex = 1;
        
    // Parse the GUID
    int iErr = _stscanf_s(pszBusGuid, REG_GUID_FORMAT_W, SVSUTIL_PGUID_ELEMENTS(&pguidBus));
    DEBUGCHK(iErr != 0 && iErr != EOF);

    //All we care about is that the szName is unique. AdvertiseInterface is 
    //thread-safe. Keep whacking away at index until we find a unique name.
    do 
    {
        if(fAdd)
        {
            pHidTouch->dwInstanceId = dwIndex++;
        }
        VERIFY(SUCCEEDED(StringCchPrintf(szName,
                        _countof(szName),
                        _T("%S_%03u"), HID_TOUCH_NAME, pHidTouch->dwInstanceId)));
        iErr = AdvertiseInterface(pguidBus, szName, fAdd);
    }while (!iErr);
    
    DEBUGMSG(ZONE_FUNCTION, (L"TCHHID:%S::InterfaceName= %S", __FUNCTION__, szName));
    return;
}
#endif

static 
DWORD
GetMaximumContactCount(
	HANDLE hDevice,
	PCHID_FUNCS pHidFuncs,
	PHIDP_PREPARSED_DATA phidpPreparsedData,
	PHIDP_CAPS pHidpCaps
	)
{
	NTSTATUS status;
	PHIDP_VALUE_CAPS pValueCaps = NULL;
	USHORT ValueCapsLength = 0;
	DWORD cbMaxHidPacket = 0;
	DWORD cbHidPacket = 0;
	PCHAR pbHidPacket = NULL;
	DWORD dwRet = 0;

	ValueCapsLength = pHidpCaps->NumberFeatureValueCaps;
	pValueCaps = (PHIDP_VALUE_CAPS)LocalAlloc(LPTR, sizeof(HIDP_VALUE_CAPS) * ValueCapsLength);
	
	status = HidP_GetValueCaps(
  		HidP_Feature,
  		pValueCaps,
  		&ValueCapsLength,
		phidpPreparsedData);
    if (status != HIDP_STATUS_SUCCESS) {
    	RETAILMSG(1, (_T("%S: GLE=0x%08x\r\n"), __FUNCTION__, GetLastError()));
		goto EXIT;
	}

	// The first byte is the report identifier.
	cbMaxHidPacket = pHidpCaps->FeatureReportByteLength + 1;
	pbHidPacket = (PCHAR)LocalAlloc(LPTR, cbMaxHidPacket);

	for (USHORT i = 0; i < ValueCapsLength; i++) {
		pbHidPacket[0] = pValueCaps[i].ReportID;
		DWORD value;

        status = pHidFuncs->lpGetReport(
            hDevice,
            HidP_Feature,
            pbHidPacket,
            cbMaxHidPacket,
            &cbHidPacket,
            INFINITE);
        if (NT_SUCCESS(status)) 
		{
			status = GetDigitizerUsageValue(HidP_Feature, 0, HID_USAGE_DIGITIZER_CONTACT_MAX, 
					phidpPreparsedData, pbHidPacket + 1, cbHidPacket, &value);
        	if (NT_SUCCESS(status)) 
			{
				dwRet = value;
			}
        }
	}

EXIT:	
	if (pValueCaps)
		LocalFree(pValueCaps);
	if (pbHidPacket)
		LocalFree(pbHidPacket);

	return dwRet;
}

static 
BOOL
SetSerialModeReport(
	HANDLE hDevice,
	PCHID_FUNCS pHidFuncs,
	PHIDP_CAPS pHidpCaps
	)
{
    DWORD dwErr;
    DWORD cbMaxHidPacket = 0;
    PCHAR pbHidPacket = NULL;
    BOOL  fRet = FALSE;

    // The first byte is the report identifier.
    cbMaxHidPacket = pHidpCaps->FeatureReportByteLength + 1;
    pbHidPacket = (PCHAR)LocalAlloc(LPTR, cbMaxHidPacket);

    ZeroMemory(pbHidPacket, cbMaxHidPacket);

    pbHidPacket[0] = 0x05; // Report ID
    pbHidPacket[1] = 0x00; // Interface No. Lower
    pbHidPacket[2] = 0x00; // Interface No. Upper

    dwErr = pHidFuncs->lpSetReport(
        hDevice,
        HidP_Feature,
        pbHidPacket,
        cbMaxHidPacket,
        INFINITE);
    if (dwErr == ERROR_SUCCESS) {
        fRet = TRUE;
    }
    else {
        DEBUGMSG(ZONE_ERROR, (_T("%S: pHidFuncs->SetReport error\r\n"), __FUNCTION__));
    }
    
    if (pbHidPacket) {
        LocalFree(pbHidPacket);
    }
	return fRet;
}

#ifdef DEBUG

static
void
ValidateHidTouch(
    PHID_TOUCH pHidTouch
    )
{    
    PREFAST_DEBUGCHK(pHidTouch != NULL);
    DEBUGCHK(pHidTouch->dwSig == HID_TOUCH_SIG);
    DEBUGCHK(pHidTouch->hDevice != NULL);
    DEBUGCHK(pHidTouch->pHidFuncs != NULL);
    DEBUGCHK(pHidTouch->phidpPreparsedData != NULL);
    DEBUGCHK(pHidTouch->hidpCaps.UsagePage == HID_USAGE_PAGE_DIGITIZER);
    DEBUGCHK(pHidTouch->hidpCaps.Usage == 4);
    if (pHidTouch->fhThreadInited == TRUE) {
        DEBUGCHK(pHidTouch->hThread != NULL);
    }
}

#endif // DEBUG

// Entry point for the HID driver. Initializes the structures for this 
// keyboard and starts the thread that will receive interrupt reports.
extern "C"
BOOL
HIDDeviceAttach(
    HID_HANDLE                 hDevice, 
    PCHID_FUNCS                pHidFuncs,
    const HID_DRIVER_SETTINGS * /*pDriverSettings*/,
    PHIDP_PREPARSED_DATA       phidpPreparsedData,
    PVOID                     *ppvNotifyParameter,
    DWORD                      /*dwUnused*/
    )
{
    BOOL fRet = FALSE;
    BOOL retSetMode;
    PHID_TOUCH pHidTouch;
    size_t cbUsages;

	DEBUGCHK(hDevice != NULL);
    DEBUGCHK(pHidFuncs != NULL);
    DEBUGCHK(phidpPreparsedData != NULL);
    PREFAST_DEBUGCHK(ppvNotifyParameter != NULL);

    pHidTouch = (PHID_TOUCH) LocalAlloc(LPTR, sizeof(HID_TOUCH));
    if (pHidTouch == NULL) {
        DEBUGMSG(ZONE_ERROR, (_T("%S: LocalAlloc error:%d\r\n"), __FUNCTION__, GetLastError()));
        goto EXIT;
    }

    pHidTouch->dwSig = HID_TOUCH_SIG;
    pHidTouch->hDevice = hDevice;
    pHidTouch->pHidFuncs = pHidFuncs;
    pHidTouch->phidpPreparsedData = phidpPreparsedData;
    HidP_GetCaps(pHidTouch->phidpPreparsedData, &pHidTouch->hidpCaps);

    pHidTouch->dwMaxUsages = HidP_MaxUsageListLength(HidP_Input, 
        						HID_USAGE_PAGE_DIGITIZER, phidpPreparsedData);
    cbUsages = pHidTouch->dwMaxUsages * sizeof(USAGE);

	if (pHidTouch->hidpCaps.UsagePage != HID_USAGE_PAGE_DIGITIZER) 
	{
		fRet =  FALSE;
		goto EXIT;
	}

	if (!pHidTouch->dwMaxUsages) 
	{
		fRet =  FALSE;
		goto EXIT;
	}

	// Change report format winXP -> win7
	retSetMode = SetSerialModeReport(hDevice, pHidTouch->pHidFuncs, &pHidTouch->hidpCaps);
	if (retSetMode == FALSE)
	{
		fRet = FALSE;
		goto EXIT;
	}

	g_HidTouchProps.dwMaxContact = GetMaximumContactCount(hDevice, pHidTouch->pHidFuncs, 
									pHidTouch->phidpPreparsedData, &pHidTouch->hidpCaps);
	if (!g_HidTouchProps.dwMaxContact) 
	{
		goto EXIT;
	}

    DEBUGMSG(ZONE_INFO, (_T("%S: dwMaxContact = %d\r\n"), __FUNCTION__, g_HidTouchProps.dwMaxContact));

	pHidTouch->puUsages = (PUSAGE)LocalAlloc(LPTR, cbUsages);
    if (pHidTouch->puUsages == NULL) {
       DEBUGMSG(ZONE_ERROR, (_T("%S: LocalAlloc error:%d\r\n"), __FUNCTION__, GetLastError()));
       goto EXIT;
    }

    // Create the message-queue that will be used to report the inputs 
	pHidTouch->hMsgQueueWriter = CreateTouchMsgQueue(FALSE);
    if (pHidTouch->hMsgQueueWriter == NULL)
	{
        goto EXIT;
    }

    // Create the thread that will receive reports from this device
    pHidTouch->hThread = CreateThread(NULL, 0, TouchThreadProc, pHidTouch, 0, NULL);
    if (pHidTouch->hThread == NULL) {
        DEBUGMSG(ZONE_ERROR, (_T("%S: Failed creating mouse thread\r\n"), 
            __FUNCTION__));
        goto EXIT;
    }
#ifdef DEBUG
    pHidTouch->fhThreadInited = TRUE;
#endif

    *ppvNotifyParameter = pHidTouch;
    ValidateHidTouch(pHidTouch);

#if 0 // delete by unitec k.kato
    // Advertise HID Touch interface attach.
    AdvertiseHidTouchInterface(pHidTouch, TRUE);
#endif

	// Indicate that the touch-panel is attached
	pHidTouch->hHidAttach = SetHidAttachEvent();

    fRet = TRUE;
EXIT:
    if ((fRet == FALSE) && (pHidTouch != NULL)) {
        FreeHidTouch(pHidTouch);
    }
    
    return fRet;
}
#ifdef DEBUG
// Match function with typedef.
static LPHID_CLIENT_ATTACH g_pfnDeviceAttach = HIDDeviceAttach;
#endif


// Entry point for the HID driver to give us notifications.
extern "C" 
BOOL 
WINAPI
HIDDeviceNotifications(
    DWORD  dwMsg,
    WPARAM wParam, // Message parameter
    PVOID  pvNotifyParameter
    )
{
    BOOL fRet = FALSE;
    PHID_TOUCH pHidTouch = (PHID_TOUCH) pvNotifyParameter;
	TCHHID_INFO info;

    UNREFERENCED_PARAMETER(wParam);

    if (VALID_HID_TOUCH(pHidTouch) == FALSE) {
        DEBUGMSG(ZONE_ERROR, (_T("%S: Received invalid structure pointer\r\n"), __FUNCTION__));
        goto EXIT;
    }

    ValidateHidTouch(pHidTouch);
    
    switch(dwMsg) {
        case HID_CLOSE_DEVICE:
            // Free all of our resources.
            WaitForSingleObject(pHidTouch->hThread, INFINITE);
            CloseHandle(pHidTouch->hThread);
            CloseHandle(pHidTouch->hHidAttach);

			// Send ups for each point that is still down.
			info.dwCount = g_HidTouchProps.dwMaxContact;
			info.pBspInput = (PTCH_BSP_INPUT_INFO)LocalAlloc(LPTR, 
					sizeof(TCH_BSP_INPUT_INFO) * info.dwCount);
			if (info.pBspInput)
			{
				for (UINT i = 0; i < info.dwCount; i++)
				{
					info.pBspInput[i].Flags = TouchSampleValidFlag;
				}
				HidTouchPutSample(pHidTouch, &info);
				LocalFree(info.pBspInput);
			}

#if 0 // delete by unitec k.kato
            //Advertise HID Touch interface detach
            AdvertiseHidTouchInterface(pHidTouch, FALSE);
#endif
            
            FreeHidTouch(pHidTouch);
            fRet = TRUE;
            break;

        default:
            DEBUGMSG(ZONE_ERROR, (_T("%S: Unhandled message %u\r\n"), __FUNCTION__));
            break;
    };

EXIT:
    return fRet;
}
#ifdef DEBUG
// Match function with typedef.
static LPHID_CLIENT_NOTIFICATIONS g_pfnDeviceNotifications = HIDDeviceNotifications;
#endif
