//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//
//
// Use of this sample source code is subject to the terms of the Microsoft
// license agreement under which you licensed this sample source code. If
// you did not accept the terms of the license agreement, you are not
// authorized to use this sample source code. For the terms of the license,
// please see the license agreement between you and Microsoft or, if applicable,
// see the LICENSE.RTF on your install media or the root of your tools installation.
// THE SAMPLE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES OR INDEMNITIES.
//

//
// Copyright (C) 2011, UNITEC Co., LTD. All Rights Reserved.
// THIS SOURCE CODE, AND ITS USE AND DISTRIBUTION, IS SUBJECT TO THE TERMS
// AND CONDITIONS OF THE APPLICABLE LICENSE AGREEMENT
//

#ifndef __TCHSTREAMPDD_H_
#define __TCHSTREAMPDD_H_

#if __cplusplus
extern "C" {
#endif

// undef here since these are already defined.
#undef ZONE_INIT
#undef ZONE_ERROR

#include <tchstream.h>
#include <tchstreamddsi.h>
#include <tchddsi.h>
#include <mtouch.h>

//------------------------------------------------------------------------------
// Sample rate during polling.
#define DEFAULT_SAMPLE_RATE_HZ              100
#define SAMPLE_RATE_LOW_HZ         			DEFAULT_SAMPLE_RATE_HZ
#define SAMPLE_RATE_HIGH_HZ       			DEFAULT_SAMPLE_RATE_HZ
#define CAL_DELTA_RESET                    	20
#define CAL_HOLD_STEADY_TIME         		1500
#define SAMPLE_RANGE_MIN_VALUE         		0
#define SAMPLE_RANGE_MAX_VALUE        		10

#define THREAD_EXIT_TIMEOUT                 3000 //3sec
#define RK_HARDWARE_DEVICEMAP_TOUCH     	(TEXT("HARDWARE\\DEVICEMAP\\TOUCH"))
#define RV_CALIBRATION_DATA             	(TEXT("CalibrationData"))

//------------------------------------------------------------------------------

#define dimof(x)            				(sizeof(x)/sizeof(x[0]))
#define offset(s, f)        				FIELD_OFFSET(s, f)
#define fieldsize(s, f)     				sizeof(((s*)0)->f)


//------------------------------------------------------------------------------
//
//  Function:  ReadRegistryParams
//
//  This function initializes driver default settings from registry based on
//  table passed as argument.
//
#define PARAM_DWORD             1
#define PARAM_STRING            2
#define PARAM_MULTIDWORD        3
#define PARAM_BIN               4


typedef struct {
    LPTSTR name;
    DWORD  type;
    BOOL   required;
    DWORD  offset;
    DWORD  size;
    PVOID  pDefault;
} DEVICE_REGISTRY_PARAM;

 DWORD GetDeviceRegistryParams(
    LPCWSTR szContext, VOID *pBase, DWORD count,
    const DEVICE_REGISTRY_PARAM params[]
);
static DWORD GetStringParam(
    HKEY hKey,
    void *pBase,
    const DEVICE_REGISTRY_PARAM *pParam
);
static DWORD GetDWordParam(
    HKEY hKey, VOID *pBase, const DEVICE_REGISTRY_PARAM *pParam
);
static DWORD GetMultiDWordParam(
    HKEY hKey, VOID *pBase, const DEVICE_REGISTRY_PARAM *pParam
);
static DWORD GetBinParam(
    HKEY hKey, VOID *pBase, const DEVICE_REGISTRY_PARAM *pParam
);


//------------------------------------------------------------------------------
// local data structures
//

typedef struct
{
    BOOL        bInitialized;
    DWORD       nSampleRate;
    DWORD       dwSysIntr;
    DWORD       dwSamplingTimeOut;
    BOOL        bTerminateIST;
    HANDLE      hTouchPanelEvent;
    DWORD       dwPowerState;
    DWORD       dwISTPriority;

	// BSP data
	TCH_BSP_INTERFACE_INFO BspIfc; 
	DWORD		hBspContext;
	DWORD		dwMaxContact;

	// IST data
    PCETOUCHINPUT pNowInputs;
    PCETOUCHINPUT pLastInputs;
	TCH_BSP_INPUT_INFO * pBspInputs;
} TOUCH_DEVICE;


//------------------------------------------------------------------------------
//  Device registry parameters
static const DEVICE_REGISTRY_PARAM s_deviceRegParams[] = {
    {
        L"SampleRate", PARAM_DWORD, FALSE, offset(TOUCH_DEVICE, nSampleRate),
        fieldsize(TOUCH_DEVICE, nSampleRate), (VOID*)DEFAULT_SAMPLE_RATE_HZ
    },
    {
        L"Priority256", PARAM_DWORD, FALSE, offset(TOUCH_DEVICE, dwISTPriority),
        fieldsize(TOUCH_DEVICE, dwISTPriority), (VOID*)DEFAULT_THREAD_PRIORITY
    },
    {
        L"SysIntr", PARAM_DWORD, FALSE, offset(TOUCH_DEVICE, dwSysIntr),
        fieldsize(TOUCH_DEVICE, dwSysIntr), (VOID*)SYSINTR_NOP
    }
};

//------------------------------------------------------------------------------
// global variables
//
static TOUCH_DEVICE g_TouchDevice =  {
    FALSE,                                          //bInitialized
    DEFAULT_SAMPLE_RATE_HZ,                         //nSampleRate
    SYSINTR_NOP,                                    //dwSysIntr
    0,                                              //dwSamplingTimeOut
    FALSE,                                          //bTerminateIST
    0,                                              //hTouchPanelEvent
    D0,                                             //dwPowerState
    DEFAULT_THREAD_PRIORITY,                        //dwISTPriority
	{0},
	0,
};

// Internal functions.
static HRESULT PDDCalibrationThread();
void PDDStartCalibrationThread();

BOOL PDDGetTouchData(UINT32 * pxPos, UINT32 * pyPos);
BOOL PDDGetRegistrySettings( PDWORD );
BOOL PDDInitializeHardware(LPCTSTR pszActiveKey );
VOID PDDDeinitializeHardware( VOID );
VOID  PDDTouchPanelDisable();
BOOL  PDDTouchPanelEnable();
ULONG PDDTouchIST(PVOID   reserved);
void PDDTouchPanelPowerHandler(BOOL boff);


#ifdef __cplusplus
}
#endif

#endif // __TCHSTREAMPDD_H_
